package net.knarcraft.bookswithoutborders;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import net.knarcraft.bookswithoutborders.utility.BookFormatter;
import net.knarcraft.bookswithoutborders.utility.EncryptionHelper;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import net.milkbowl.vault.economy.Economy;
import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.OfflinePlayer;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.command.ConsoleCommandSender;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemFactory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.inventory.meta.ItemMeta;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.plugin.java.JavaPlugin;

public class BooksWithoutBorders extends JavaPlugin {
    protected static int bookDuplicateLimit;
    //The separating string between the book title and the book author
    protected static String titleAuthorSeparator;
    protected static String loreSep;
    protected static final String SLASH = System.getProperty("file.separator");
    protected static List<String> firstBooks;
    protected static String welcomeMessage;
    protected static List<String> existingPlayers;
    protected static Economy eco;
    protected static Material bookPriceType = null;
    protected static double bookPriceQuantity;
    protected static boolean authorOnlyCopy, useYml, adminDecrypt;
    protected static ItemFactory iF;
    protected static Map<String, List<String>> loadList;
    protected static BooksWithoutBorders bwb;
    protected static BooksWithoutBordersListener bL;
    protected static ConsoleCommandSender consoleSender;
    protected static String bookFolder;
    protected static ChatColor errorColor = ChatColor.RED;
    protected static ChatColor successColor = ChatColor.GREEN;
    protected static ChatColor commandColor = ChatColor.YELLOW;

    @Override
    public void onEnable() {
        bwb = this;
        consoleSender = this.getServer().getConsoleSender();
        bookFolder = this.getDataFolder().getAbsolutePath() + SLASH + "Books" + SLASH;
        bL = new BooksWithoutBordersListener();
        loadList = new HashMap<>();
        firstBooks = new ArrayList<>();

        if (SLASH != null && init()) {
            this.getServer().getPluginManager().registerEvents(bL, this);
        } else {
            this.getPluginLoader().disablePlugin(this);
        }
    }

    protected boolean init() {
        //initializes Item Factory
        try {
            iF = this.getServer().getItemFactory();
        } catch (java.lang.NoSuchMethodError nsmE) {
            sendErrorMessage(consoleSender, "Warning! [BooksWithoutBorders] failed to initialize!");
            sendErrorMessage(consoleSender, "Please confirm the correct version of [BooksWithoutBorders] is");
            sendErrorMessage(consoleSender, "being run for this version of bukkit!");
            return false;
        }

        //Pulls config values
        if (!loadConfig()) {
            return false;
        }

        this.getConfig().set("Options.Save_Books_in_Yaml_Format", useYml);
        this.getConfig().set("Options.Max_Number_of_Duplicates", bookDuplicateLimit);
        this.getConfig().set("Options.Title-Author_Separator", titleAuthorSeparator);
        this.getConfig().set("Options.Lore_line_separator", loreSep);
        this.getConfig().set("Options.Books_for_new_players", firstBooks);
        this.getConfig().set("Options.Message_for_new_players", welcomeMessage);
        if (bookPriceType != null) {
            if (bookPriceType != Material.AIR)
                this.getConfig().set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
            else
                this.getConfig().set("Options.Price_to_create_book.Item_type", "Economy");
        } else
            this.getConfig().set("Options.Price_to_create_book.Item_type", "Item type name");
        this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
        this.getConfig().set("Options.Admin_Auto_Decrypt", adminDecrypt);
        this.getConfig().set("Options.Author_Only_Copy", authorOnlyCopy);

        //Handles old book and quill settings
        if (this.getConfig().contains("Options.Require_book_and_quill_to_create_book")) {
            sendSuccessMessage(consoleSender, "[BooksWithoutBorders] Found old config setting \"Require_book_and_quill_to_create_book\"");
            sendSuccessMessage(consoleSender, "Updating to \"Price_to_create_book\" settings");

            if (this.getConfig().getBoolean("Options.Require_book_and_quill_to_create_book")) {
                bookPriceType = Material.WRITABLE_BOOK;
                bookPriceQuantity = 1;
                this.getConfig().set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
                this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
            }

            this.getConfig().set("Options.Require_book_and_quill_to_create_book", null);
        }

        this.saveConfig();

        File fTest = new File(bookFolder);
        File efTest = new File(bookFolder + "Encrypted" + SLASH);

        if (!fTest.exists()) {
            try {
                if (!fTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        if (!efTest.exists()) {
            try {
                if (!efTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }

        return loadExistingPlayers();
    }

    protected boolean loadConfig() {
        this.reloadConfig();
        try {
            useYml = this.getConfig().getBoolean("Options.Save_Books_in_Yaml_Format", true);
            bookDuplicateLimit = this.getConfig().getInt("Options.Max_Number_of_Duplicates", 5);
            titleAuthorSeparator = this.getConfig().getString("Options.Title-Author_Separator", ",");
            loreSep = this.getConfig().getString("Options.Lore_line_separator", "~");
            adminDecrypt = this.getConfig().getBoolean("Options.Admin_Auto_Decrypt", false);
            authorOnlyCopy = this.getConfig().getBoolean("Options.Author_Only_Copy", false);

            firstBooks = this.getConfig().getStringList("Options.Books_for_new_players");
            if (this.getConfig().contains("Options.Book_for_new_players"))
                firstBooks.add(this.getConfig().getString("Options.Book_for_new_players"));
            if (firstBooks.isEmpty())
                firstBooks.add(" ");

            welcomeMessage = this.getConfig().getString("Options.Message_for_new_players", " ");

            //Converts string into material
            String sMaterial = this.getConfig().getString("Options.Price_to_create_book.Item_type", " ");
            if (sMaterial.equalsIgnoreCase("Economy")) {
                if (setupEconomy())
                    bookPriceType = Material.AIR;
                else {
                    sendErrorMessage(consoleSender, "BooksWithoutBorders failed to hook into Vault! Book price not set!");
                    bookPriceType = null;
                }
            } else if (!sMaterial.equalsIgnoreCase(" ")) {
                for (Material m : Material.values()) {
                    if (m.toString().equalsIgnoreCase(sMaterial))
                        bookPriceType = m;
                }
            }
            bookPriceQuantity = this.getConfig().getDouble("Options.Price_to_create_book.Required_quantity", 0);

            //makes sure titleAuthorSeparator is a valid value
            titleAuthorSeparator = cleanString(titleAuthorSeparator);
            if (titleAuthorSeparator.length() != 1) {
                sendErrorMessage(consoleSender, "Title-Author_Separator is set to an invalid value!");
                sendErrorMessage(consoleSender, "Reverting to default value of \",\"");
                titleAuthorSeparator = ",";
                this.getConfig().set("Options.Title-Author_Separator", titleAuthorSeparator);
            }
        } catch (Exception e) {
            sendErrorMessage(consoleSender, "Warning! Config.yml failed to load!");
            sendErrorMessage(consoleSender, "Try Looking for settings that are missing values!");
            return false;
        }

        return true;
    }

    private boolean setupEconomy() {
        Plugin plugin = getServer().getPluginManager().getPlugin("Vault");

        if (plugin != null) {
            RegisteredServiceProvider<Economy> economyProvider = getServer().getServicesManager().getRegistration(net.milkbowl.vault.economy.Economy.class);
            if (economyProvider != null) {
                eco = economyProvider.getProvider();
            }
        }

        return (eco != null);
    }

    protected boolean loadExistingPlayers() {
        File fTest = new File(this.getDataFolder().getAbsolutePath() + SLASH + "Existing Players.txt");
        existingPlayers = new ArrayList<>();

        if (!fTest.exists()) {
            try {
                if (!fTest.createNewFile()) {
                    sendErrorMessage(consoleSender, "Could not create Existing Players file! Aborting...");
                    return false;
                }
                PrintWriter pw = new PrintWriter(new FileWriter(fTest));
                OfflinePlayer[] players = this.getServer().getOfflinePlayers();

                for (OfflinePlayer player : players) {
                    existingPlayers.add(player.getName());
                }
                for (String player : existingPlayers) {
                    pw.println(player);
                }
                pw.close();
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        } else {
            try {
                BufferedReader br = new BufferedReader(new FileReader(fTest));
                String nullcheck = br.readLine();

                while (nullcheck != null) {
                    existingPlayers.add(nullcheck);
                    nullcheck = br.readLine();
                }
                br.close();
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }

        return true;
    }

    protected void addExistingPlayer(String playerName) {
        existingPlayers.add(playerName);

        try {
            PrintWriter pw = new PrintWriter(new FileWriter(this.getDataFolder().getAbsolutePath() + SLASH + "Existing Players.txt", true));
            pw.println(playerName);
            pw.close();
        } catch (IOException e) {
            sendErrorMessage(consoleSender, "Failed to add " + playerName + " to Existing Players list");
        }
    }

    protected boolean hasPlayedBefore(String playerName) {
        if (!existingPlayers.contains(playerName)) {
            addExistingPlayer(playerName);
            return false;
        }

        return true;
    }

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {

        if ((command.getName().equalsIgnoreCase("BooksWithoutBorders") || command.getName().equalsIgnoreCase("bwb"))) {
            if (args.length == 0 && sender instanceof Player) {
                //lists commands
                sender.sendMessage(commandColor + "Use: /bwb [Command]");
                sender.sendMessage(commandColor + "[] denote parameters");

                if (booksHavePrice()) {
                    if (bookPriceType != Material.AIR) {
                        sendErrorMessage(sender, "[" + (int) bookPriceQuantity + " " + bookPriceType.toString() + "(s) are required to create a book]");
                    } else {
                        sendErrorMessage(sender, "[" + BooksWithoutBorders.eco.format(bookPriceQuantity) + " is required to create a book]");
                    }
                }
                sender.sendMessage(commandColor + "Commands:");

                if (sender.hasPermission("bookswithoutborders.load")) {
                    sender.sendMessage(commandColor + "\nLoad [file name or number] [# of copies] [true/false]:");
                    sendSuccessMessage(sender, "Creates a book from the specified file and gives it to the player");
                    sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
                    sendSuccessMessage(sender, "If true is specified the book will be signed, if false it will be");
                    sendSuccessMessage(sender, "unsigned");
                }
                if (sender.hasPermission("bookswithoutborders.loadpublic")) {
                    sender.sendMessage(commandColor + "loadPublic [file name or number] [# of copies] [true/false]:");
                    sendSuccessMessage(sender, "Same as Load, but views files in the public directory");
                }
                if (sender.hasPermission("bookswithoutborders.save")) {
                    sender.sendMessage("\n" + commandColor + "Save [true/false]: " + successColor + "Saves the book the player is");
                    sendSuccessMessage(sender, "holding to a text file in a private directory");
                    sendSuccessMessage(sender, "If true is specified, a book of the same name by the same");
                    sendSuccessMessage(sender, "author will be overwritten by the new book");
                }
                if (sender.hasPermission("bookswithoutborders.savepublic")) {
                    sender.sendMessage(commandColor + "savePublic [true/false]: " + successColor + "Same as Save,");
                    sendSuccessMessage(sender, "but saves files in the public directory");
                }
                if (sender.hasPermission("bookswithoutborders.give")) {
                    sender.sendMessage("\n" + commandColor + "Give [file name or number] [playername] [# of copies] [true/false]:");
                    sendSuccessMessage(sender, "Gives the selected player a book from your personal directory");
                }
                if (sender.hasPermission("bookswithoutborders.givepublic")) {
                    sender.sendMessage(commandColor + "givePublic [file name or number] [playername] [# of copies] [true/false]:");
                    sendSuccessMessage(sender, "Same as give, but uses books from the public directory");
                }
                if (sender.hasPermission("bookswithoutborders.delete")) {
                    sender.sendMessage(commandColor + "\nDelete [file name or number]: " + successColor + "Deletes the specified");
                    sendSuccessMessage(sender, "file in the player's directory");
                    sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
                }
                if (sender.hasPermission("bookswithoutborders.admin")) {
                    sender.sendMessage(commandColor + "deletePublic [file name or number]: " + successColor + "Same as Delete,");
                    sendSuccessMessage(sender, "but deletes files in the public directory");
                    sender.sendMessage(commandColor + "\nReload:" + successColor + " Reloads BwB's configuration file");
                }

                if (sender.hasPermission("bookswithoutborders.unsign"))
                    sender.sendMessage("\n" + commandColor + "Unsign: " + successColor + "Unsigns the book the player is holding");
                if (sender.hasPermission("bookswithoutborders.copy"))
                    sender.sendMessage("\n" + commandColor + "Copy [number of copies]: " + successColor + "Copies the book the player is holding");

                if (sender.hasPermission("bookswithoutborders.encrypt")) {
                    sender.sendMessage("\n" + commandColor + "Encrypt [key] [style]: " + successColor + "Encrypts the book the player is holding");
                    sender.sendMessage(commandColor + "[key]" + successColor + " is required and can be any phrase or number excluding spaces");
                    sender.sendMessage(commandColor + "[style]" + successColor + " is not required. Possible values are \"DNA\" or \"Magic\"");
                }
                if (sender.hasPermission("bookswithoutborders.groupencrypt")) {
                    sender.sendMessage("\n" + commandColor + "groupEncrypt [group name] [key] [style]: " + successColor + "Encrypts book so that only players with the" +
                            "\n bookswithoutborders.decrypt." + commandColor + "[group name]" + successColor + " permission may decrypt the book by holding and left clicking the book");
                }
                if (sender.hasPermission("bookswithoutborders.decrypt")) {
                    sender.sendMessage("\n" + commandColor + "Decrypt [key]: " + successColor + "Decrypts the book the player is holding");
                    sender.sendMessage(commandColor + "[key]" + successColor + " is required and MUST be IDENTICAL to the key used to encrypt held book");
                }
                if (sender.hasPermission("bookswithoutborders.settitle")) {
                    sender.sendMessage("\n" + commandColor + "setTitle [title]: " + successColor + "Sets the title of the book/item the player is holding");
                }
                if (sender.hasPermission("bookswithoutborders.setauthor")) {
                    sender.sendMessage("\n" + commandColor + "setAuthor [author]: " + successColor + "Sets the author of the book the player is holding");
                }
                if (sender.hasPermission("bookswithoutborders.setlore")) {
                    sender.sendMessage("\n" + commandColor + "setLore [lore]: " + successColor + "Sets the lore of the item the player is holding");
                    sendSuccessMessage(sender, "Insert the lore_line_separator character to force a new line\n[\"~\" by default]");
                }
                if (sender.hasPermission("bookswithoutborders.setbookprice"))
                    sender.sendMessage("\n" + commandColor + "setBookPrice [Item/Eco] [quantity]: " + successColor + "Sets the per-book-price to create a book via commands." +
                            "\nIf [Item], the item in the player's hand in the amount of [quanity] will be the price.\nIf [Eco], a Vault based economy will be used for price." +
                            "\nIf neither [Item/Eco] or [quantity] are specified the current price to create books will be removed.");
                return true;
            }

            if (args.length == 0) {
                //lists commands from console
                sender.sendMessage(commandColor + "Use: /bwb [Command]");
                sender.sendMessage(commandColor + "[] denote parameters");
                sender.sendMessage(commandColor + "Commands:");
                sender.sendMessage(commandColor + "\nReload:" + successColor + " Reloads BwB's config file");
                sender.sendMessage(commandColor + "givePublic [file name or number] [playername] [true/false]: " + successColor);
                sendSuccessMessage(sender, "Gives the selected player a book from the public directory");
                sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");
                sender.sendMessage(commandColor + "deletePublic [file name or number]: " + successColor + "Deletes the specified");
                sendSuccessMessage(sender, "file in the public directory");
                sendSuccessMessage(sender, "If no file is specified, a list of available files is returned");

                return true;
            }

            if (args[0].equalsIgnoreCase("reload")) {
                if (sender instanceof Player) {
                    if (!sender.hasPermission("bookswithoutborders.admin") && !sender.hasPermission("*")) {
                        sendErrorMessage(sender, "You don't have permission to use this command!");
                        return false;
                    }
                }

                if (loadConfig() && loadExistingPlayers()) {
                    sendSuccessMessage(sender, "BooksWithoutBorders configuration reloaded!");
                } else {
                    sendErrorMessage(sender, "Reload Failed!");
                    sendErrorMessage(sender, "See console for details");
                }
                return true;
            }

            //Player only commands
            if (sender instanceof Player player) {

                if (args[0].equalsIgnoreCase("savePublic")) {
                    if (!sender.hasPermission("bookswithoutborders.savepublic")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (player.getItemInHand().getType() == Material.WRITTEN_BOOK || player.getItemInHand().getType() == Material.WRITABLE_BOOK) {
                        if (args.length == 2) {
                            saveBook(player, args[1], true);
                        } else {
                            saveBook((Player) sender, "false", true);
                        }
                        return true;
                    } else {
                        sendErrorMessage(sender, "You must be holding a written book or book and quill to save it!");
                        return false;
                    }
                }

                if (args[0].equalsIgnoreCase("save")) {
                    if (!sender.hasPermission("bookswithoutborders.save")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK || ((Player) sender).getItemInHand().getType() == Material.WRITABLE_BOOK) {
                        if (args.length == 2) {
                            saveBook((Player) sender, args[1], false);
                        } else {
                            saveBook((Player) sender, "false", false);
                        }
                        return true;
                    } else {
                        sendErrorMessage(sender, "You must be holding a written book or book and quill to save it!");
                        return false;
                    }
                }

                if (args[0].equalsIgnoreCase("loadPublic")) {
                    if (!sender.hasPermission("bookswithoutborders.loadpublic")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (((Player) sender).getInventory().firstEmpty() == -1) {
                        sendErrorMessage(sender, "You must have space in your inventory to load books!");
                        return false;
                    }

                    if (args.length == 1) {
                        loadList.put(sender.getName(), listFiles(sender, true, 0, false));
                        return true;
                    }

                    for (int x = 0; x < args[1].length(); x++) {
                        if (!Character.isDigit(args[1].charAt(x)))
                            break;
                        if (x == args[1].length() - 1)
                            loadList.put(sender.getName(), listFiles(sender, true, 0, true));
                    }

                    ItemStack newBook;
                    try {
                        if (args.length == 4)
                            newBook = loadBook(sender, cleanString(args[1]), args[3], "public", Integer.parseInt(args[2]));
                        else if (args.length == 3) {
                            if (args[2].equalsIgnoreCase("true") || args[2].equalsIgnoreCase("false"))
                                newBook = loadBook(sender, cleanString(args[1]), args[2], "public");
                            else
                                newBook = loadBook(sender, cleanString(args[1]), "true", "public", Integer.parseInt(args[2]));
                        } else
                            newBook = loadBook(sender, cleanString(args[1]), "true", "public");

                        if (newBook != null) {
                            ((Player) sender).getInventory().addItem(newBook);
                            sendSuccessMessage(sender, "Book created!");
                            return true;
                        } else {
                            sendErrorMessage(sender, "Book failed to load!");
                            return false;
                        }
                    } catch (NumberFormatException e) {
                        sendErrorMessage(sender, "Invalid number of book copies specified!");
                        return false;
                    }
                }

                if (args[0].equalsIgnoreCase("load")) {
                    if (!sender.hasPermission("bookswithoutborders.load")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (((Player) sender).getInventory().firstEmpty() == -1) {
                        sendErrorMessage(sender, "You must have space in your inventory to load books!");
                        return false;
                    }

                    if (args.length == 1) {
                        loadList.put(sender.getName(), listFiles(sender, false, 0, false));
                        return true;
                    }

                    for (int x = 0; x < args[1].length(); x++) {
                        if (!Character.isDigit(args[1].charAt(x)))
                            break;
                        if (x == args[1].length() - 1)
                            loadList.put(sender.getName(), listFiles(sender, false, 0, true));
                    }

                    ItemStack newBook;
                    try {

                        if (args.length == 4)
                            newBook = loadBook(sender, cleanString(args[1]), args[3], "player", Integer.parseInt(args[2]));
                        else if (args.length == 3) {
                            if (args[2].equalsIgnoreCase("true") || args[2].equalsIgnoreCase("false"))
                                newBook = loadBook(sender, cleanString(args[1]), args[2], "player");
                            else
                                newBook = loadBook(sender, cleanString(args[1]), "true", "player", Integer.parseInt(args[2]));
                        } else
                            newBook = loadBook(sender, cleanString(args[1]), "true", "player");

                        if (newBook != null) {
                            ((Player) sender).getInventory().addItem(newBook);
                            sendSuccessMessage(sender, "Book created!");
                            return true;
                        } else {
                            sendErrorMessage(sender, "Book failed to load!");
                            return false;
                        }
                    } catch (NumberFormatException e) {
                        sendErrorMessage(sender, "Invalid number of book copies specified!");
                        return false;
                    }
                }

                if (args[0].equalsIgnoreCase("give")) {
                    return commandGive(sender, args);
                }

                if (args[0].equalsIgnoreCase("delete")) {
                    return commandDelete(sender, args);
                }

                if (args[0].equalsIgnoreCase("unsign")) {
                    if (!sender.hasPermission("bookswithoutborders.unsign")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (!(((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK)) {
                        sendErrorMessage(sender, "You must be holding a written book to unsign it!");
                        return false;
                    }

                    boolean hasBookInMainHand = player.getInventory().getItemInMainHand().getType() == Material.WRITTEN_BOOK;
                    boolean hasBookInOffHand = player.getInventory().getItemInOffHand().getType() == Material.WRITTEN_BOOK;

                    //Find which hand the player is using to hold the book. If holding one in each, throw an error
                    if (hasBookInMainHand && hasBookInOffHand) {
                        sendErrorMessage(sender, "You cannot un-sign two books at once. Please un-equip one " +
                                "of the books you're holding!");
                    } else if (hasBookInMainHand) {
                        unSignHeldBook(player, true);
                    } else if (hasBookInOffHand) {
                        unSignHeldBook(player, false);
                    }
                    return true;
                }

                if (args[0].equalsIgnoreCase("copy")) {
                    if (!sender.hasPermission("bookswithoutborders.copy")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (!(((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK)) {
                        sendErrorMessage(sender, "You must be holding a written book to copy it!");
                        return false;
                    }

                    if (args.length < 2) {
                        sendErrorMessage(sender, "You must specifiy the number of copies to be made!");
                        return false;
                    }

                    try {
                        if (Integer.parseInt(args[1]) > 0) {
                            if (authorOnlyCopy && !sender.hasPermission("bookswithoutborders.bypassauthoronlycopy")) {
                                if (!isAuthor((Player) sender, (BookMeta) ((Player) sender).getItemInHand().getItemMeta()))
                                    return false;
                            }

                            if (booksHavePrice() &&
                                    !sender.hasPermission("bookswithoutborders.bypassbookprice") &&
                                    cannotPayForBookPrinting((Player) sender, Integer.parseInt(args[1]))) {
                                return false;
                            }

                            ((Player) sender).getItemInHand().setAmount(((Player) sender).getItemInHand().getAmount() + Integer.parseInt(args[1]));
                            sendSuccessMessage(sender, "Book copied!");
                        } else {
                            sendErrorMessage(sender, "Book not copied!");
                            sendErrorMessage(sender, "Number specified was invalid!");
                            return false;
                        }
                    } catch (NumberFormatException e) {
                        sendErrorMessage(sender, "Book not copied!");
                        sendErrorMessage(sender, "Number specified was invalid!");
                        return false;
                    }
                    return true;
                }

                if (args[0].equalsIgnoreCase("encrypt")) {
                    if (!sender.hasPermission("bookswithoutborders.encrypt")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (!(((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK)) {
                        sendErrorMessage(sender, "You must be holding a written book to encrypt it!");
                        return false;
                    }

                    if (args.length < 2) {
                        sendErrorMessage(sender, "You must specify a key to encrypt a book!");
                        return false;
                    }
                    if (args.length > 3) {
                        sendErrorMessage(sender, "Too many command options specified!");
                        return false;
                    }

                    if (!((BookMeta) ((Player) sender).getItemInHand().getItemMeta()).hasPages()) {
                        sendErrorMessage(sender, "Book must have contents to encrypt!");
                        return false;
                    }

                    ItemStack eBook;

                    if (args.length == 3) {
                        eBook = encryptBook((Player) sender, true, args[1], EncryptionStyle.getFromString(args[2]), "");
                    } else {
                        eBook = encryptBook((Player) sender, true, args[1], EncryptionStyle.SUBSTITUTION, "");
                    }

                    if (eBook != null) {
                        ((Player) sender).setItemInHand(eBook);
                        return true;
                    } else
                        return false;
                }

                if (args[0].equalsIgnoreCase("groupEncrypt")) {
                    if (!sender.hasPermission("bookswithoutborders.groupencrypt")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (!(((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK)) {
                        sendErrorMessage(sender, "You must be holding a written book to encrypt it!");
                        return false;
                    }

                    if (args.length < 3) {
                        sendErrorMessage(sender, "You must specify a key to encrypt a book!");
                        return false;
                    }
                    if (args.length > 4) {
                        sendErrorMessage(sender, "Too many command options specified!");
                        return false;
                    }

                    if (!((BookMeta) ((Player) sender).getItemInHand().getItemMeta()).hasPages()) {
                        sendErrorMessage(sender, "Book must have contents to encrypt!");
                        return false;
                    }

                    if (((Player) sender).getItemInHand().getItemMeta().hasLore()
                            && ((Player) sender).getItemInHand().getItemMeta().getLore().get(0).contains(" encrypted]")) {
                        sendErrorMessage(sender, "Book is already group encrypted!");
                        return false;
                    }

                    ItemStack eBook;

                    if (args.length == 4) {
                        eBook = encryptBook((Player) sender, true, args[2], EncryptionStyle.getFromString(args[3]), args[1]);
                    } else {
                        eBook = encryptBook((Player) sender, true, args[2], EncryptionStyle.SUBSTITUTION, args[1]);
                    }

                    if (eBook != null) {
                        ((Player) sender).setItemInHand(eBook);
                        return true;
                    } else
                        return false;
                }

                if (args[0].equalsIgnoreCase("decrypt")) {
                    if (!sender.hasPermission("bookswithoutborders.decrypt")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (!(((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK)) {
                        sendErrorMessage(sender, "You must be holding a written book to decrypt it!");
                        return false;
                    }

                    if (args.length == 1 && adminDecrypt && sender.hasPermission("bookswithoutborders.admin") && sender instanceof Player) {
                        String path = bookFolder + "Encrypted" + SLASH;
                        String fname = (!((BookMeta) ((Player) sender).getItemInHand().getItemMeta()).hasTitle()) ? "Untitled," + sender.getName() :
                                ((BookMeta) ((Player) sender).getItemInHand().getItemMeta()).getTitle() + titleAuthorSeparator + ((BookMeta) ((Player) sender).getItemInHand().getItemMeta()).getAuthor();
                        File eDir = new File(path);

                        String key = "";
                        for (int i = 0; i < eDir.list().length; i++) {
                            if (eDir.list()[i].contains(fname)) {
                                key = eDir.list()[i].substring(eDir.list()[i].indexOf("[") + 1, eDir.list()[i].indexOf("]"));
                                break;
                            }
                        }

                        if (!key.equalsIgnoreCase("")) {
                            ItemStack book = eLoad(((Player) sender), key, false);
                            if (book != null) {
                                ((Player) sender).setItemInHand(book);
                                sendSuccessMessage(sender, "Book auto-decrypted!");
                                return true;
                            } else
                                return false;
                        } else {
                            sendErrorMessage(sender, "No matching encrypted book found!");
                            return false;
                        }
                    } else if (args.length == 1) {
                        sendErrorMessage(sender, "No decryption password given!");
                        return false;
                    }

                    String key = EncryptionHelper.getNumberKeyFromStringKey(args[1]);

                    ItemStack book = eLoad(((Player) sender), key, true);
                    if (book != null) {
                        ((Player) sender).setItemInHand(book);
                        sendSuccessMessage(sender, "Book decrypted!");
                        return true;
                    } else
                        return false;
                }

                if (args[0].equalsIgnoreCase("setTitle")) {
                    if (!sender.hasPermission("bookswithoutborders.settitle")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (args.length < 2) {
                        sendErrorMessage(sender, "Too few command arguments!");
                        return false;
                    }

                    if (((Player) sender).getItemInHand().getType() == Material.AIR) {
                        sendErrorMessage(sender, "You must be holding an item to set title!");
                        return false;
                    }

                    StringBuilder title = new StringBuilder();
                    for (int x = 1; x < args.length; x++) {
                        if (x == 1)
                            title.append(args[x]);
                        else
                            title.append(" ").append(args[x]);
                    }
                    title = new StringBuilder(ChatColor.translateAlternateColorCodes('&', title.toString()));

                    ItemMeta meta;
                    if (((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK) {
                        BookMeta bMeta = (BookMeta) ((Player) sender).getItemInHand().getItemMeta();
                        bMeta.setTitle(title.toString());
                        meta = bMeta;
                    } else {
                        ItemMeta iMeta = ((Player) sender).getItemInHand().getItemMeta();
                        iMeta.setDisplayName(title.toString());
                        meta = iMeta;
                    }

                    ((Player) sender).getItemInHand().setItemMeta(meta);
                    sendSuccessMessage(sender, "Title set to " + title + "!");
                    return true;
                }

                if (args[0].equalsIgnoreCase("setAuthor")) {
                    if (!sender.hasPermission("bookswithoutborders.setauthor")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (args.length < 2) {
                        sendErrorMessage(sender, "Too few command arguments!");
                        return false;
                    }

                    if (!(((Player) sender).getItemInHand().getType() == Material.WRITTEN_BOOK)) {
                        sendErrorMessage(sender, "You must be holding a written book to set author!");
                        return false;
                    }

                    BookMeta meta = (BookMeta) ((Player) sender).getItemInHand().getItemMeta();
                    StringBuilder author = new StringBuilder();
                    for (int x = 1; x < args.length; x++) {
                        if (x == 1)
                            author.append(args[x]);
                        else
                            author.append(" ").append(args[x]);
                    }
                    meta.setAuthor(author.toString());
                    ((Player) sender).getItemInHand().setItemMeta(meta);
                    sendSuccessMessage(sender, "Book author set to " + author + "!");
                    return true;
                }

                if (args[0].equalsIgnoreCase("setLore")) {
                    if (!sender.hasPermission("bookswithoutborders.setlore")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }

                    if (args.length < 2) {
                        sendErrorMessage(sender, "Missing a command argument!");
                        return false;
                    }

                    if (((Player) sender).getItemInHand().getType() == Material.AIR) {
                        sendErrorMessage(sender, "Must be holding an item to set lore!");
                        return false;
                    }

                    StringBuilder rawlore = new StringBuilder();
                    for (int x = 1; x < args.length; x++) {
                        if (x == 1)
                            rawlore.append(args[x]);
                        else
                            rawlore.append(" ").append(args[x]);
                    }

                    rawlore = new StringBuilder(ChatColor.translateAlternateColorCodes('&', rawlore.toString()));
                    String[] lores = rawlore.toString().split(loreSep);
                    List<String> nulore = new ArrayList<>(Arrays.asList(lores));

                    ItemMeta meta = ((Player) sender).getItemInHand().getItemMeta();
                    meta.setLore(nulore);
                    ((Player) sender).getItemInHand().setItemMeta(meta);
                    sendSuccessMessage(sender, "Added lore to item!");
                    return true;
                }
            }

            if (args[0].equalsIgnoreCase("givePublic")) {
                if (sender instanceof Player) {
                    if (!sender.hasPermission("bookswithoutborders.givepublic")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }
                }

                if (args.length == 2 || args.length > 5) {
                    sendErrorMessage(sender, "Incorrect number of arguments for this command!");
                    return false;
                }

                if (args.length == 1) {
                    loadList.put(sender.getName(), listFiles(sender, true, 0, false));
                    return true;
                }

                for (int x = 0; x < args[1].length(); x++) {
                    if (!Character.isDigit(args[1].charAt(x)))
                        break;
                    if (x == args[1].length() - 1)
                        loadList.put(sender.getName(), listFiles(sender, true, 0, true));
                }

                ItemStack newBook;
                Player receivingPlayer = this.getServer().getPlayer(args[2]);
                if (receivingPlayer == null) {
                    sendErrorMessage(sender, "Player not found!");
                    return false;
                }

                if (receivingPlayer.getInventory().firstEmpty() == -1) {
                    sendErrorMessage(sender, "Receiving player must have space in their inventory to receive books!");
                    return false;
                }
                //bwb give [bookname] [player] [numCopies] [issigned]
                try {
                    if (args.length == 5)
                        newBook = loadBook(sender, cleanString(args[1]), args[4], "public", Integer.parseInt(args[3]));
                    else if (args.length == 4) {
                        if (args[3].equalsIgnoreCase("true") || args[3].equalsIgnoreCase("false"))
                            newBook = loadBook(sender, cleanString(args[1]), args[3], "public");
                        else
                            newBook = loadBook(sender, cleanString(args[1]), "true", "public", Integer.parseInt(args[3]));
                    } else
                        newBook = loadBook(sender, cleanString(args[1]), "true", "public");

                    if (newBook != null) {
                        receivingPlayer.getInventory().addItem(newBook);
                        sendSuccessMessage(sender, "Book sent!");
                        sendSuccessMessage(receivingPlayer, "Book received!");
                        return true;
                    } else {
                        sendErrorMessage(sender, "Book failed to load!");
                        return false;
                    }
                } catch (NumberFormatException e) {
                    sendErrorMessage(sender, "Invalid number of book copies specified!");
                    return false;
                }
            }

            if (args[0].equalsIgnoreCase("deletepublic")) {

                if (sender instanceof Player) {
                    if (!sender.hasPermission("bookswithoutborders.admin")) {
                        sendErrorMessage(sender, " You don't have permission to use this command!");
                        return false;
                    }
                }

                //lists delete-able files
                if (args.length == 1) {
                    loadList.put(sender.getName(), listFiles(sender, true, 0, false));
                    return true;
                }
                //actual deletion
                if (args.length == 2) {
                    if (!loadList.containsKey(sender.getName())) {
                        sendErrorMessage(sender, "You must first use /bwb deletePublic to create a list of delete-able files!");
                        return false;
                    } else {
                        if (!loadList.get(sender.getName()).isEmpty()) {
                            deleteBook((sender), args[1], true);
                            return true;
                        } else {
                            sendErrorMessage(sender, "No files available to delete!");
                            return false;
                        }
                    }
                }
            }

            if (args[0].equalsIgnoreCase("setBookPrice")) {
                if (!sender.hasPermission("bookswithoutborders.setbookprice")) {
                    sendErrorMessage(sender, " You don't have permission to use this command!");
                    return false;
                }

                if (args.length == 1) {
                    bookPriceType = null;
                    bookPriceQuantity = 0;
                    this.getConfig().set("Options.Price_to_create_book.Item_type", "Item type name");
                    this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
                    this.saveConfig();

                    sendSuccessMessage(sender, "Price to create books removed!");
                    return true;
                }

                if (args.length < 3) {
                    sendErrorMessage(sender, "[Item/Eco] and [quantity] must be specified!");
                    return false;
                }

                if (!args[1].equalsIgnoreCase("Item") && !args[1].equalsIgnoreCase("Eco")) {
                    sendErrorMessage(sender, "Price type must be \"Item\" or \"Eco\"!");
                    return false;
                }

                try {
                    if (Double.parseDouble(args[2]) <= 0) {
                        sendErrorMessage(sender, "[quantity] must be greater than 0!");
                        return false;
                    }

                    if (args[1].equalsIgnoreCase("Item")) {
                        if (((Player) sender).getItemInHand().getType() == Material.AIR) {
                            sendErrorMessage(sender, "Must be holding an item to set an [Item] price!");
                            return false;
                        }

                        bookPriceType = ((Player) sender).getItemInHand().getType();
                        bookPriceQuantity = Double.parseDouble(args[2]);
                        this.getConfig().set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
                        this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
                        this.saveConfig();

                        sendSuccessMessage(sender, "Book creation price set to " + (int) bookPriceQuantity + " " + bookPriceType.toString() + "(s)!");
                        return true;
                    } else if (args[1].equalsIgnoreCase("Eco")) {
                        if (setupEconomy()) {
                            bookPriceQuantity = Double.parseDouble(args[2]);
                            bookPriceType = Material.AIR;
                            this.getConfig().set("Options.Price_to_create_book.Item_type", "Economy");
                            this.getConfig().set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
                            this.saveConfig();

                            sendSuccessMessage(sender, "Book creation price set to " + eco.format(bookPriceQuantity) + "!");
                            return true;
                        } else {
                            sendErrorMessage(sender, "BooksWithoutBorders failed to hook into Vault! Book price not set!");
                            return false;
                        }

                    }
                } catch (NumberFormatException e) {
                    sendErrorMessage(sender, "[quantity] must be a number!");
                }
            }

            sendErrorMessage(sender, "Command not recognized! Use " + commandColor + "/BwB" + errorColor + " for more info!");
        }
        return false;
    }

    /**
     * Executes the delete command
     * @param sender <p>The command sender which sent the command</p>
     * @param args <p>The arguments given</p>
     * @return <p>True if the command was executed successfully</p>
     */
    protected boolean commandDelete(CommandSender sender, String[] args) {
        if (!sender.hasPermission("bookswithoutborders.delete")) {
            sendErrorMessage(sender, " You don't have permission to use this command!");
            return false;
        }

        //List deletable files
        if (args.length == 1) {
            loadList.put(sender.getName(), listFiles(sender, false, 0, false));
            return true;
        }
        //Delete the file
        if (args.length == 2) {
            if (!loadList.containsKey(sender.getName())) {
                sendErrorMessage(sender, "You must first use /bwb delete to create a list of delete-able files!");
                return false;
            } else {
                if (!loadList.get(sender.getName()).isEmpty()) {
                    deleteBook(sender, args[1], false);
                    return true;
                } else {
                    sendErrorMessage(sender, "No files available to delete!");
                    return false;
                }
            }
        }
        sendErrorMessage(sender, "Incorrect number of arguments for this command!");
        return false;
    }

    /**
     * Executes the give command
     * @param sender <p>The command sender which sent the command</p>
     * @param args <p>The arguments given</p>
     * @return <p>True if the command was executed successfully</p>
     */
    protected boolean commandGive(CommandSender sender, String[] args) {
        if (!sender.hasPermission("bookswithoutborders.give")) {
            sendErrorMessage(sender, " You don't have permission to use this command!");
            return false;
        }

        if (args.length == 2 || args.length > 5) {
            sendErrorMessage(sender, "Incorrect number of arguments for this command!");
            return false;
        }

        if (args.length == 1) {
            loadList.put(sender.getName(), listFiles(sender, false, 0, false));
            return true;
        }

        for (int x = 0; x < args[1].length(); x++) {
            if (!Character.isDigit(args[1].charAt(x)))
                break;
            if (x == args[1].length() - 1)
                loadList.put(sender.getName(), listFiles(sender, false, 0, true));
        }

        ItemStack newBook;
        Player receivingPlayer = this.getServer().getPlayer(args[2]);
        if (receivingPlayer == null) {
            sendErrorMessage(sender, "Player not found!");
            return false;
        }

        if (receivingPlayer.getInventory().firstEmpty() == -1) {
            sendErrorMessage(sender, "Receiving player must have space in their inventory to receive books!");
            return false;
        }

        //bwb give [bookname] [player] [numCopies] [issigned]
        try {

            if (args.length == 5)
                newBook = loadBook(sender, cleanString(args[1]), args[4], "player", Integer.parseInt(args[3]));
            else if (args.length == 4) {
                if (args[3].equalsIgnoreCase("true") || args[3].equalsIgnoreCase("false"))
                    newBook = loadBook(sender, cleanString(args[1]), args[3], "player");
                else
                    newBook = loadBook(sender, cleanString(args[1]), "true", "player", Integer.parseInt(args[3]));
            } else
                newBook = loadBook(sender, cleanString(args[1]), "true", "player");

            if (newBook != null) {
                receivingPlayer.getInventory().addItem(newBook);
                sendSuccessMessage(sender, "Book sent!");
                sendSuccessMessage(receivingPlayer, "Book received!");
                return true;
            } else {
                sendErrorMessage(sender, "Book failed to load!");
                return false;
            }
        } catch (NumberFormatException e) {
            sendErrorMessage(sender, "Invalid number of book copies specified!");
            return false;
        }
    }

    protected String cleanString(String fname) {
        //removes illegal characters
        if (fname.contains("/")) {
            fname = fname.replace("/", "");
        }
        if (fname.contains("\\")) {
            fname = fname.replace("\\", "");
        }
        if (fname.contains("*")) {
            fname = fname.replace("*", "");
        }
        if (fname.contains(":")) {
            fname = fname.replace(":", "");
        }
        if (fname.contains("|")) {
            fname = fname.replace("|", "");
        }
        if (fname.contains("<")) {
            fname = fname.replace("<", "");
        }
        if (fname.contains(">")) {
            fname = fname.replace(">", "");
        }
        if (fname.contains("?")) {
            fname = fname.replace("?", "");
        }
        if (fname.contains("\"")) {
            fname = fname.replace("\"", "");
        }
        return fname;
    }

    protected String fixName(String fname, Boolean isLoad) {
        if (isLoad) {
            fname = fname.replace("_", " ");
        } else {
            fname = fname.replace(" ", "_");
        }
        return fname;
    }

    protected void bookToYml(String path, String fname, BookMeta book) throws IOException {
        FileConfiguration bookYml = YamlConfiguration.loadConfiguration(new File(path, "blank"));

        if (book.hasTitle())
            bookYml.set("Title", book.getTitle());
        if (book.hasAuthor())
            bookYml.set("Author", book.getAuthor());
        if (book.hasPages())
            bookYml.set("Pages", book.getPages());
        if (book.hasLore())
            bookYml.set("Lore", book.getLore());

        bookYml.save(path + fname + ".yml");
    }

    protected BookMeta bookFromYml(File file, BookMeta bDat) {
        //Handles old style loading
        if (file.getName().substring(file.getName().lastIndexOf(".")).equals(".txt"))
            bDat = bookFromTXT(file.getName(), file, bDat);
        else {
            try {
                FileConfiguration bookYml = YamlConfiguration.loadConfiguration(file);

                bDat.setTitle(bookYml.getString("Title", "Untitled"));
                bDat.setAuthor(bookYml.getString("Author", "Unknown"));
                bDat.setPages(bookYml.getStringList("Pages"));
                bDat.setLore(bookYml.getStringList("Lore"));
            } catch (IllegalArgumentException e) {
                return null;
            }
        }

        return bDat;
    }

    protected void bookToTXT(String path, String fname, BookMeta book) throws IOException {
        FileWriter fw = new FileWriter(path + fname + ".txt");
        PrintWriter pw = new PrintWriter(fw);
        List<String> pages = book.getPages();

        pw.println("[Book]");
        for (String page : pages) {
            pw.println(page);
        }
        pw.close();
    }

    protected BookMeta bookFromTXT(String fname, File file, BookMeta bDat) {
        String author;
        String title;
        if (fname.contains(titleAuthorSeparator)) {
            author = fname.substring(fname.indexOf(titleAuthorSeparator) + 1, fname.length() - 4);
            title = fname.substring(0, fname.indexOf(titleAuthorSeparator));
        } else {
            author = "Unknown";
            title = fname.substring(0, fname.length() - 4);
        }

        title = fixName(title, true);

        List<String> rawPages = new ArrayList<>();
        String nullcheck;

        try {
            FileReader fileReader = new FileReader(file);
            BufferedReader bufferedReader = new BufferedReader(fileReader);

            rawPages.add(bufferedReader.readLine());

            if (rawPages.get(0) == null) {
                bufferedReader.close();
                return null;
            }

            //if a book is being loaded its content need not be adjusted
            if (rawPages.get(0).equalsIgnoreCase("[Book]")) {
                rawPages.remove(0);
                nullcheck = bufferedReader.readLine();
                while (nullcheck != null) {
                    rawPages.add(nullcheck);
                    nullcheck = bufferedReader.readLine();
                }
            } else {
                //adjusts content to page length
                while (rawPages.get(rawPages.size() - 1) != null) {
                    BookFormatter.formatLastPage(rawPages);
                    rawPages.add(bufferedReader.readLine());
                }
            }
            bufferedReader.close();
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }

        List<String> pages = new ArrayList<>(cleanList(rawPages));

        bDat.setAuthor(author);
        bDat.setTitle(title);
        bDat.setPages(pages);

        return bDat;
    }

    protected void saveBook(Player player, String dupe, Boolean pub) {
        BookMeta book = (BookMeta) player.getItemInHand().getItemMeta();
        String path;
        if (pub) {
            path = bookFolder;
        } else {
            path = bookFolder + cleanString(player.getName()) + SLASH;
        }
        String fname;
        if (!book.hasTitle()) {
            fname = "Untitled," + player.getName();
        } else {
            fname = book.getTitle() + titleAuthorSeparator + book.getAuthor();
        }
        fname = cleanString(fname);
        fname = fixName(fname, false);
        int dupes = 0;

        //checks to see if the file name is already taken and adjusts it if so
        File file = new File(path);

        //if this is a private save, make sure the dir exists
        if (!file.exists()) {
            if (!file.mkdir()) {
                sendErrorMessage(player, "Saving Failed! If this continues to happen, consult server admin!");
                return;
            }
        }

        if (file.listFiles().length > 0) {
            File[] fl = file.listFiles();
            for (int x = 0; x < fl.length; x++) {
                if (dupes == 0) {
                    if (fl[x].getName().equalsIgnoreCase(fname)) {
                        dupes++;
                        x = -1;
                    }
                } else {
                    if (fl[x].getName().equalsIgnoreCase("(" + dupes + ")" + fname)) {
                        dupes++;
                        x = -1;
                    }
                }
            }
        }
        if (dupes > 0) {
            if (!fname.contains("Untitled") && dupe.equalsIgnoreCase("false")) {
                sendErrorMessage(player, "Book is already saved!");
                sendErrorMessage(player, "Use " + commandColor + "/bwb Save true " + errorColor + "to overwrite!");
                return;
            }

            if (dupes > bookDuplicateLimit) {
                sendErrorMessage(player, "Maximum amount of " + fname + " duplicates reached!");
                sendErrorMessage(player, "Use " + commandColor + "/bwb Save true " + errorColor + "to overwrite!");
                return;
            }

            if (fname.contains("Untitled") && dupe.equalsIgnoreCase("false")) {
                fname = "(" + dupes + ")" + fname;
            }
        }

        try {
            if (useYml)
                bookToYml(path, fname, book);
            else
                bookToTXT(path, fname, book);

            sendSuccessMessage(player, "Book Saved as \"" + fname + "\"");
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    protected ItemStack loadBook(CommandSender sender, String fname, String isSigned, String dir) {
        return loadBook(sender, fname, isSigned, dir, 1);
    }

    protected ItemStack loadBook(CommandSender sender, String fname, String isSigned, String dir, int numCopies) {
        //checks if player is using list number to load
        for (int x = 0; x < fname.length(); x++) {
            if (!Character.isDigit(fname.charAt(x))) {
                break;
            }
            if (x == fname.length() - 1 && Character.isDigit(fname.charAt(x)) && loadList.containsKey(sender.getName())) {
                if (Integer.parseInt(fname) <= loadList.get(sender.getName()).size()) {
                    fname = loadList.get(sender.getName()).get(Integer.parseInt(fname) - 1);
                }
            }
        }

        File file;
        NameCheck:
        {
            //Extension is already present
            if (fname.lastIndexOf(".") != -1) {
                if (fname.substring(fname.lastIndexOf(".")).equals(".yml") || fname.substring(fname.lastIndexOf(".")).equals(".txt")) {
                    if (dir.equalsIgnoreCase("public"))
                        file = new File(bookFolder + fname);

                    else if (dir.equalsIgnoreCase("player"))
                        file = new File(bookFolder + cleanString(sender.getName()) + SLASH + fname);

                    else if (!dir.equalsIgnoreCase(""))
                        file = new File(bookFolder + "Encrypted" + SLASH + dir + SLASH + fname);

                    else
                        file = null;

                    if (!file.isFile()) {
                        sendErrorMessage(sender, "Incorrect file name!");
                        return null;
                    }
                    break NameCheck;
                }
            }
            //No extension present
            if (dir.equalsIgnoreCase("public"))
                file = new File(bookFolder + fname + ".yml");

            else if (dir.equalsIgnoreCase("player"))
                file = new File(bookFolder + cleanString(sender.getName()) + SLASH + fname + ".yml");

            else if (!dir.equalsIgnoreCase(""))
                file = new File(bookFolder + "Encrypted" + SLASH + dir + SLASH + fname + ".yml");

            else
                file = null;

            if (!file.isFile()) {
                if (dir.equalsIgnoreCase("public"))
                    file = new File(bookFolder + fname + ".txt");

                else if (dir.equalsIgnoreCase("player"))
                    file = new File(bookFolder + cleanString(sender.getName()) + SLASH + fname + ".txt");

                else if (!dir.equalsIgnoreCase(""))
                    file = new File(bookFolder + "Encrypted" + SLASH + dir + SLASH + fname + ".txt");

                else
                    file = null;

                if (!file.isFile()) {
                    sendErrorMessage(sender, "Incorrect file name!");
                    return null;
                }
            }
        }

        ItemStack book;
        BookMeta bDat = (BookMeta) iF.getItemMeta(Material.WRITTEN_BOOK);
        if (isSigned.equalsIgnoreCase("true")) {
            book = new ItemStack(Material.WRITTEN_BOOK);
        } else {
            book = new ItemStack(Material.WRITABLE_BOOK);
        }

        bookFromYml(file, bDat);
        if (bDat == null) {
            sendErrorMessage(sender, "File was blank!!");
            return null;
        }

        if (!dir.equalsIgnoreCase("public") && !dir.equalsIgnoreCase("player") && !dir.equalsIgnoreCase("") && bDat.hasLore()) {
            List<String> newLore = new ArrayList<>(bDat.getLore());
            newLore.remove(0);
            bDat.setLore(newLore);
        }

        book.setItemMeta(bDat);
        book.setAmount(numCopies);

        if (booksHavePrice() && !sender.hasPermission("bookswithoutborders.bypassbookprice") &&
                (dir.equalsIgnoreCase("public") || dir.equalsIgnoreCase("player") ||
                        dir.equalsIgnoreCase("")) &&
                cannotPayForBookPrinting((Player) sender, numCopies)) {
            return null;
        }

        return book;
    }

    protected Boolean eSave(Player player, BookMeta book, String key) {
        String path = bookFolder + "Encrypted" + SLASH;
        String fname = (!book.hasTitle()) ? "Untitled," + player.getName() :
                book.getTitle() + titleAuthorSeparator + book.getAuthor();

        fname = "[" + key + "]" + fname;
        fname = cleanString(fname);
        fname = fixName(fname, false);

        //cancels saving if file is already encrypted
        File file = (useYml) ? new File(path + fname + ".yml") : new File(path + fname + ".txt");
        if (file.isFile()) {
            return true;
        }

        try {
            bookToYml(path, fname, book);
        } catch (IOException e) {
            e.printStackTrace();
            sendErrorMessage(player, "Encryption failed!");
            return false;
        }
        return true;
    }

    protected BookMeta groupESave(Player player, BookMeta book, String key, String groupName) {
        String path = bookFolder + "Encrypted" + SLASH + cleanString(groupName) + SLASH;
        File dirTest = new File(path);
        //Creates group dir
        if (!dirTest.exists()) {
            try {
                if (!dirTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Unable to create encryption group folder!");
                    return null;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        }
        //Creates file
        String fname = (!book.hasTitle()) ? "Untitled," + player.getName() :
                book.getTitle() + titleAuthorSeparator + book.getAuthor();

        fname = cleanString(fname);
        fname = fixName(fname, false);

        List<String> nuLore = new ArrayList<>();
        nuLore.add(ChatColor.GRAY + "[" + groupName + " encrypted]");
        if (book.hasLore())
            nuLore.addAll(book.getLore());
        book.setLore(nuLore);

        //Save file
        File file = (useYml) ? new File(path + fname + ".yml") : new File(path + fname + ".txt");
        if (!file.isFile()) {
            try {
                bookToYml(path, fname, book);
            } catch (IOException e) {
                e.printStackTrace();
                sendErrorMessage(player, "Group encrypted failed!");
                return null;
            }
        }

        return book;
    }

    protected ItemStack eLoad(Player player, String key, boolean deleteEncryptedFile) {
        BookMeta book = (BookMeta) player.getItemInHand().getItemMeta();
        String path = bookFolder + "Encrypted" + SLASH;

        String fname = (!book.hasTitle()) ? "Untitled," + player.getName() : book.getTitle() + titleAuthorSeparator + book.getAuthor();
        fname = "[" + key + "]" + fname;
        fname = cleanString(fname);
        fname = fixName(fname, false);

        File file = new File(path + fname + ".yml");
        if (!file.isFile()) {
            file = new File(path + fname + ".txt");

            if (!file.isFile()) {
                sendErrorMessage(player, "Incorrect decryption key!");
                return null;
            }
        } else {
            try {
                book = bookFromYml(file, book);
            } catch (Exception e) {
                sendErrorMessage(player, "Decryption failed!");
                return null;
            }
        }

        if (deleteEncryptedFile) {
            try {
                file.delete();
            } catch (Exception e) {
                sendErrorMessage(consoleSender, "Book encryption data failed to delete upon decryption!");
                sendErrorMessage(consoleSender, "File location:" + file.getPath());
            }
        }

        ItemStack nubook = new ItemStack(Material.WRITTEN_BOOK);//Book(book.getAuthor(), book.getTitle(), pages, 1, 387);
        nubook.setItemMeta(book);
        nubook.setAmount(player.getItemInHand().getAmount());
        return nubook;
    }

    protected List<String> listFiles(CommandSender sender, Boolean pub, int offset, boolean silent) {
        List<String> fileList = new ArrayList<>();
        File file = (pub) ? new File(bookFolder) :
                new File(bookFolder + cleanString(sender.getName()) + SLASH);
        File[] fl = file.listFiles();

        if (!file.exists()) {
            sendErrorMessage(sender, "No books have been saved!");
            return null;
        }

        if (fl.length == 0) {
            sendErrorMessage(sender, "No books have been saved!");
            return null;
        }

        if (!silent) {
            sendSuccessMessage(sender, "Available Books:");
        }
        for (File value : fl) {
            if (value.isFile()) {
                fileList.add(value.getName());
                if (!silent)
                    sender.sendMessage(ChatColor.GRAY + "[" + (fileList.size() + offset) + "] " + value.getName()/*.substring(0, fl[x].getName().length()-4)*/);
            }
        }

        return fileList;
    }

    /**
     * Removes null and empty items from a list
     *
     * @param list <p>The list to clean</p>
     * @return <p>A clean list containing all relevant values</p>
     */
    protected List<String> cleanList(List<String> list) {
        List<String> resultList = new ArrayList<>(list);
        resultList.removeIf((item) -> item == null || item.trim().isEmpty());
        return resultList;
    }

    /**
     * Encrypts a book
     * @param player <p>The player encrypting the book</p>
     * @param mainHand <p>Whether the player is holding the book in its main hand</p>
     * @param key <p>The key/password to use for encryption</p>
     * @param style <p>The encryption style to use</p>
     * @return <p>An encrypted version of the book</p>
     */
    protected ItemStack encryptBook(Player player, boolean mainHand, String key, EncryptionStyle style) {
        return encryptBook(player, mainHand, key, style, "");
    }

    /**
     * Encrypts a book
     * @param player <p>The player encrypting the book</p>
     * @param mainHand <p>Whether the player is holding the book in its main hand</p>
     * @param key <p>The key/password to use for encryption</p>
     * @param style <p>The encryption style to use</p>
     * @param groupName <p>The name of the group to encrypt for, or "" otherwise</p>
     * @return <p>An encrypted version of the book</p>
     */
    protected ItemStack encryptBook(Player player, boolean mainHand, String key, EncryptionStyle style, String groupName) {
        //converts user supplied key into integer form
        String integerKey = EncryptionHelper.getNumberKeyFromStringKey(key);

        BookMeta book = getHeldBookMetadata(player, mainHand);

        if (!book.hasPages()) {
            sendErrorMessage(player, "Book is empty!");
            return null;
        }

        //Save the book's un-encrypted contents to a file
        BookMeta newMetadata = saveBookPlaintext(groupName, player, book, integerKey);
        if (newMetadata == null) {
            return null;
        }

        //Get the encrypted pages
        List<String> encryptedPages = EncryptionHelper.encryptBookPages(book, style, integerKey, player);
        if (encryptedPages == null) {
            return null;
        }

        //Format the last page just in case
        BookFormatter.formatLastPage(encryptedPages);
        //Remove empty pages
        List<String> newPages = cleanList(encryptedPages);

        ItemStack encryptedBook = createEncryptedBook(book, newPages, mainHand, player, newMetadata);

        sendSuccessMessage(player, "Book encrypted!");
        return encryptedBook;
    }

    /**
     * Creates a new encrypted book
     * @param book <p>The book to encrypt</p>
     * @param newPages <p>The new encrypted pages</p>
     * @param mainHand <p>Whether the book is held in the player's main hand or off hand</p>
     * @param player <p>The player encrypting the book</p>
     * @param newMetadata <p>The new metadata of the book</p>
     * @return <p>An encrypted version of the book</p>
     */
    private ItemStack createEncryptedBook(BookMeta book, List<String> newPages, boolean mainHand, Player player, BookMeta newMetadata) {
        //Create the encrypted book
        ItemStack encryptedBook = new ItemStack(Material.WRITTEN_BOOK);
        book.setPages(newPages);
        encryptedBook.setItemMeta(book);

        //Update item amount
        if (mainHand) {
            encryptedBook.setAmount(player.getInventory().getItemInMainHand().getAmount());
        } else {
            encryptedBook.setAmount(player.getInventory().getItemInOffHand().getAmount());
        }
        //Set new item metadata
        encryptedBook.setItemMeta(newMetadata);

        return encryptedBook;
    }

    /**
     * Saves a book's plain text to a file
     * @param groupName <p>The group who's allowed to decrypt the book, or ""</p>
     * @param player <p>The player trying to encrypt the book</p>
     * @param book <p>The book to encrypt</p>
     * @param integerKey <p>The key used to encrypt the book</p>
     * @return <p>The new metadata for the book, or null if it could not be saved</p>
     */
    private BookMeta saveBookPlaintext(String groupName, Player player, BookMeta book, String integerKey) {
        BookMeta newMetadata = book;
        boolean wasSaved;
        if (groupName.trim().isEmpty()) {
            wasSaved = eSave(player, book, integerKey);
        } else {
            newMetadata = groupESave(player, book, integerKey, groupName);
            wasSaved = newMetadata != null;
        }
        if (wasSaved) {
            return newMetadata;
        } else {
            return null;
        }
    }

    /**
     * Sends a success message to a command sender (player or a console)
     * @param sender <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    private void sendSuccessMessage(CommandSender sender, String message) {
        sender.sendMessage(successColor + message);
    }

    /**
     * Sends an error message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendErrorMessage(CommandSender sender, String message) {
        sender.sendMessage(errorColor + message);
    }

    /**
     * Deletes the given book
     *
     * @param sender   <p>The sender of the command to delete the book</p>
     * @param fileName <p>The file name of the book</p>
     * @param isPublic <p>Whether the book to delete is public or not</p>
     */
    protected void deleteBook(CommandSender sender, String fileName, Boolean isPublic) {
        //If the file name is an index of the load list, load the book
        try {
            int loadListIndex = Integer.parseInt(fileName);
            String senderName = sender.getName();
            if (loadList.containsKey(senderName) && loadListIndex <= loadList.get(senderName).size()) {
                fileName = loadList.get(senderName).get(loadListIndex - 1);
            }
        } catch (NumberFormatException ignored) {
        }

        //Get the file to be deleted
        File file;
        if (isPublic) {
            file = FileHelper.getBookFile(bookFolder + fileName);
        } else {
            file = FileHelper.getBookFile(bookFolder + cleanString(sender.getName()) + SLASH + fileName);
        }

        //Send message if no such file could be found
        if (file == null) {
            sendErrorMessage(sender, "Incorrect file name!");
            return;
        }

        //Try to delete the file
        try {
            if (file.delete()) {
                sendSuccessMessage(sender, "\"" + fileName + "\" deleted successfully");
            } else {
                sendErrorMessage(sender, "Deletion failed without an exception!");
            }
        } catch (Exception e) {
            sendErrorMessage(sender, "Deletion failed!");
        }
    }

    /**
     * Unsigns the player's currently held book
     *
     * @param player   <p>The player holding the book</p>
     * @param mainHand <p>Whether the player is holding the book in its main hand or its off hand</p>
     */
    protected void unSignHeldBook(Player player, boolean mainHand) {
        //Get the old book
        BookMeta oldBook = getHeldBookMetadata(player, mainHand);

        //UnSign the book
        ItemStack newBook = new ItemStack(Material.WRITABLE_BOOK);
        newBook.setItemMeta(oldBook);

        replaceHeldBook(player, newBook, mainHand);
    }

    /**
     * Gets metadata about the player's held book
     *
     * @param player   <p>The player holding the book</p>
     * @param mainHand <p>Whether to get information about a book in the player's main hand or off hand</p>
     * @return <p>Information about the held book</p>
     */
    private BookMeta getHeldBookMetadata(Player player, boolean mainHand) {
        if (mainHand) {
            return (BookMeta) player.getInventory().getItemInMainHand().getItemMeta();
        } else {
            return (BookMeta) player.getInventory().getItemInOffHand().getItemMeta();
        }
    }

    /**
     * Replaces the player's held item
     *
     * @param player   <p>The player to replace the item for</p>
     * @param newBook  <p>The new book the player should hold</p>
     * @param mainHand <p>Whether to replace the item in the player's main hand or off hand</p>
     */
    private void replaceHeldBook(Player player, ItemStack newBook, boolean mainHand) {
        if (mainHand) {
            player.getInventory().setItemInMainHand(newBook);
        } else {
            player.getInventory().setItemInOffHand(newBook);
        }
    }

    /**
     * Checks whether books have a price for printing them
     *
     * @return <p>True if players need to pay for printing books</p>
     */
    protected boolean booksHavePrice() {
        return (bookPriceType != null && bookPriceQuantity > 0);
    }

    /**
     * Makes the player pay for printing a given number of books
     *
     * @param player    <p>The player printing the books</p>
     * @param numCopies <p>The number of copies the player is trying to print</p>
     * @return <p>True if the player cannot pay for the printing of the books</p>
     */
    protected boolean cannotPayForBookPrinting(Player player, int numCopies) {
        //BookPriceQuantity: How many items are required to pay for each book
        //BookPriceType: Which item is used to pay for the books. AIR = use economy
        double cost = bookPriceQuantity * numCopies;
        int itemCost = (int) cost;

        if (bookPriceType == Material.AIR) {
            return !payForBookPrintingEconomy(player, cost, numCopies);
        } else {
            if (player.getInventory().contains(bookPriceType, itemCost)) {
                payForBookPrintingItem(player, itemCost);
                return false;
            } else {
                sendErrorMessage(player, itemCost + " " + bookPriceType.toString() +
                        "(s) are required for this command!");
                return true;
            }
        }
    }

    /**
     * Takes payment for printing a number of books by withdrawing the correct item
     *
     * @param player   <p>The player which needs to pay</p>
     * @param itemCost <p>The number of items to pay</p>
     */
    private void payForBookPrintingItem(Player player, int itemCost) {
        PlayerInventory playerInventory = player.getInventory();

        int clearedAmount = 0;
        while (clearedAmount < itemCost) {
            int firstItemIndex = playerInventory.first(bookPriceType);
            ItemStack firstItem = playerInventory.getItem(firstItemIndex);

            if (Objects.requireNonNull(firstItem).getAmount() <= itemCost - clearedAmount) {
                clearedAmount += firstItem.getAmount();
                player.getInventory().clear(firstItemIndex);
            } else {
                clearedAmount = itemCost;
                firstItem.setAmount(firstItem.getAmount() - (clearedAmount));
            }
        }
    }

    /**
     * Uses economy to take payment for printing a number of books
     *
     * @param player    <p>The player which needs to pay</p>
     * @param cost      <p>The cost of the book printing</p>
     * @param numCopies <p>The number of books the player is printing</p>
     * @return <p>True if the player had the money and it has been withdrawn</p>
     */
    private boolean payForBookPrintingEconomy(Player player, double cost, int numCopies) {
        Economy economy = BooksWithoutBorders.eco;
        if ((economy.getBalance(player) - cost) >= 0) {
            economy.withdrawPlayer(player, cost);
            sendSuccessMessage(player, economy.format(cost) + " withdrawn to create " + numCopies + " book(s)");
            sendSuccessMessage(player, "New balance: " + economy.format(economy.getBalance(player)));
            return true;
        } else {
            sendErrorMessage(player, economy.format(cost) + " is required for this command!");
            return false;
        }
    }

    /**
     * Checks whether the given player is the author of a given book
     *
     * @param player <p>The player to check</p>
     * @param book   <p>The book to check</p>
     * @return <p>True if the player is the book's author</p>
     */
    protected boolean isAuthor(Player player, BookMeta book) {
        String author = book.getAuthor();
        if (author != null && cleanString(player.getName()).equalsIgnoreCase(cleanString(author))) {
            return true;
        }

        sendErrorMessage(player, "You must be the author of this book to use this command!");
        return false;
    }
}
